use std::time::Duration;

use leptos::*;
use serde::{Deserialize, Serialize};

use super::backend_settings::BackendSettings;
use crate::app::components::FormControl;

#[derive(PartialEq, Debug, Clone, Serialize, Deserialize, sqlx::FromRow)]
pub struct BackendProcess {
    id: i64,
    #[sqlx(flatten)]
    settings: BackendSettings,
    status: ProcessStatus,
}

/// The `ProcessStatus` enum represents the various states a process can be in. It includes the following states:
/// - `WaitingForStart`: The process is waiting to be started.
/// - `Running`: The process is currently running.
/// - `WaitingForStop`: The process is waiting to be stopped.
/// - `Finished`: The process has completed its execution successfully.
/// - `Failed`: The process has failed or encountered an error during execution.
///
/// This enum is used to keep track of the state of a process in a backend application, allowing for proper management and control over the process lifecycle.
#[derive(
    Default, PartialEq, Debug, Clone, Serialize, Deserialize, sqlx::Type, strum::EnumString,
)]
#[sqlx(rename_all = "snake_case")]
#[strum(serialize_all = "snake_case")]
pub enum ProcessStatus {
    #[default]
    WaitingForStart,
    Running,
    WaitingForStop,
    Finished,
    Failed,
}

/// CHecks the db every 100ms to see if stuff has to be started
#[cfg(feature = "ssr")]
pub async fn run_starter_task(pool: sqlx::SqlitePool) {
    use tokio::{
        time::{self, Instant},
        try_join,
    };
    use tokio_stream::{wrappers::IntervalStream, StreamExt};

    use crate::server::backends::BackendService;
    let _ = tracing::debug_span!("starter_task");

    tracing::debug!("Starter task started");

    let service_handle = BackendService::new();
    let mut stream = IntervalStream::new(time::interval(Duration::from_millis(1000)));
    while let Some(instant) = stream.next().await {
        tracing::debug!("fire; instant={:?}", instant);

        let waiting_to_start: Vec<BackendProcess> = sqlx::query_as(
            "SELECT id, status,
                se.version, se.model_dir, se.default_model, se.created_at
            FROM backend_process INNER JOIN settings AS se
            ON backend_process.settings_version = se.version
            WHERE status == ?",
        )
        .bind(ProcessStatus::WaitingForStart)
        .fetch_all(&pool)
        .await
        .unwrap();

        let waiting_to_stop: Vec<BackendProcess> = sqlx::query_as(
            "SELECT id, status,
                se.version, se.model_dir, se.default_model, se.created_at
            FROM backend_process INNER JOIN settings AS se
            ON backend_process.settings_version = se.version
            WHERE status == ?",
        )
        .bind(ProcessStatus::WaitingForStop)
        .fetch_all(&pool)
        .await
        .unwrap();

        tracing::debug!(?waiting_to_start, ?waiting_to_stop);
        use futures::future::FutureExt;

        for to_stop in waiting_to_stop.into_iter() {
            tracing::debug!(?to_stop);
            let update_query_fut = sqlx::query!(
                "UPDATE backend_process
                SET status == ?
                WHERE id == ?",
                ProcessStatus::Finished,
                to_stop.id
            )
            .execute(&pool)
            .map(|f| f.map_err(|_e| anyhow::Error::msg("update fail")));
            let stop_fut = service_handle.stop();

            tracing::debug!(msg = "Stopping", ?to_stop);
            try_join!(update_query_fut, stop_fut).expect("stop fail");
            tracing::debug!(msg = "Stopped!", ?to_stop);
        }

        for to_start in waiting_to_start.into_iter() {
            // TODO Pass args to backend
            // TODO Save Services tehrefore allow multiples
            tracing::debug!(?to_start);

            let update_query_fut = sqlx::query!(
                "UPDATE backend_process
                SET status == ?
                WHERE id == ?",
                ProcessStatus::Running,
                to_start.id
            )
            .execute(&pool)
            .map(|f| f.map_err(|_e| anyhow::Error::msg("update fail")));

            tracing::debug!(msg = "Starting", ?to_start);
            let start_fut = service_handle.start();
            try_join!(update_query_fut, start_fut).expect("start fail");
            tracing::debug!(msg = "Started!", ?to_start);
        }
        tracing::debug!(msg="Starter task iteration finished", iteration_dur=?Instant::now().duration_since(instant))
    }
}

/// Starts a new process and updates the process status.
#[server]
pub async fn start_process() -> Result<(), ServerFnError> {
    use crate::server::pool;

    let pool = pool()?;

    // Set all existing processes to Finished
    sqlx::query!(
        "UPDATE backend_process SET status = 'waiting_for_stop' WHERE status != 'finished'"
    )
    .execute(&pool)
    .await?;

    // Get the current newest settings version
    let settings_version =
        sqlx::query_scalar!("SELECT version FROM settings ORDER BY version DESC LIMIT 1")
            .fetch_one(&pool)
            .await?;

    // Create a new ProcessStatus with the current newest settings version and initial status of waiting
    sqlx::query!(
        "INSERT INTO backend_process (settings_version, status) VALUES (?, 'waiting_for_start')",
        settings_version
    )
    .execute(&pool)
    .await?;

    Ok(())
}

/// Stops all processes and updates the process status.
#[server]
pub async fn stop_process() -> Result<(), ServerFnError> {
    use crate::server::pool;

    let pool = pool()?;

    // Set all processes to finished
    sqlx::query!("UPDATE backend_process SET status = 'waiting_for_stop'")
        .execute(&pool)
        .await?;

    Ok(())
}

/// Fetches the status of the newest process.
#[server(ProcessStatusFn)]
pub async fn process_status() -> Result<ProcessStatus, ServerFnError> {
    use crate::server::pool;

    let pool = pool()?;

    let status: Option<ProcessStatus> =
        sqlx::query_scalar!("SELECT status FROM backend_process ORDER BY id DESC LIMIT 1")
            .fetch_optional(&pool)
            .await?
            .map(|str| str.parse())
            .transpose()?;

    match status {
        Some(status) => Ok(status),
        None => Ok(ProcessStatus::Finished),
    }
}
