use std::time::Duration;

use leptos::*;
use serde::{Deserialize, Serialize};

use crate::{
    api::{process_status, ProcessStatus, StartProcess, StopProcess},
    app::components::FormControl,
};

/// A form component to start and stop the process.
#[component]
pub fn ProcessForm() -> impl IntoView {
    let gen =
        use_context::<RwSignal<super::RefreshGen>>().expect("to have found the setter provided");
    let _ = gloo_timers::callback::Interval::new(1000, move || {
        gen.update(|gen| *gen = crate::app::RefreshGen(gen.0 + 1));
    })
    .forget();

    let status = create_local_resource(gen, |_| async move {
        process_status().await.unwrap_or(ProcessStatus::Finished)
    });

    let is_running = move || match status.get().unwrap_or(ProcessStatus::Finished) {
        ProcessStatus::WaitingForStart => true,
        ProcessStatus::Running => true,
        ProcessStatus::WaitingForStop => false,
        ProcessStatus::Finished => false,
        ProcessStatus::Failed => false,
    };

    let toggle_color = move || match status.get().unwrap_or(ProcessStatus::Finished) {
        ProcessStatus::WaitingForStart => "toggle-info",
        ProcessStatus::Running => "toggle-success",
        ProcessStatus::WaitingForStop => "toggle-warning",
        ProcessStatus::Finished => "toggle-info",
        ProcessStatus::Failed => "toggle-error",
    };

    let stop_process_action = create_server_action::<StopProcess>();
    let start_process_action = create_server_action::<StartProcess>();

    let on_toggle = move |_| {
        if is_running() {
            stop_process_action.dispatch(StopProcess {});
        } else {
            start_process_action.dispatch(StartProcess {});
        }
        gen.update(|gen| *gen = crate::app::RefreshGen(gen.0 + 1));
    };

    let status_text = move || {
        if let Some(status) = status.get() {
            match status {
                ProcessStatus::WaitingForStart => "waiting to start",
                ProcessStatus::Running => "running",
                ProcessStatus::WaitingForStop => "waiting to stop",
                ProcessStatus::Finished => "finished",
                ProcessStatus::Failed => "failed",
            }
        } else {
            "Loading"
        }
    };

    view! {
        <FormControl label="Running">
            <input
                type="checkbox"
                checked=is_running
                on:click=on_toggle
                class=move || format!("toggle {}", toggle_color())
            />
        </FormControl>
        <FormControl label="Status">

            <span class="label-text">{status_text}</span>
        </FormControl>
    }
}
