use std::ffi::OsStr;

use serde::{Deserialize, Serialize};
use tokio::process::Command;

pub struct RunnerArgs {
    ctx_size:   i64,
    gpu_layers: i64,
    model_path: String,
}

impl RunnerArgs {
    pub fn new(ctx_size: i64, gpu_layers: i64, model_path: String) -> Self {
        Self {
            ctx_size,
            gpu_layers,
            model_path,
        }
    }
}

impl From<RunnerArgs> for Vec<String> {
    fn from(value: RunnerArgs) -> Self {
        vec![
            "--n-gpu-layers".to_string(),
            value.gpu_layers.to_string(),
            "--ctx-size".to_string(),
            value.ctx_size.to_string(),
            // "--host".to_string(),
            // "0.0.0.0".to_string(),
            "-m".to_string(),
            value.model_path,
        ]
    }
}

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct Runner {
    pwd:  Option<String>,
    cmd:  String,
    args: Vec<String>,
}

impl Runner {
    // FIXME does not exit properly when it is killed
    pub fn new_llamafile_bin(runner_args: RunnerArgs) -> Self {
        Self {
            pwd:  None,
            cmd:  "bash".to_string(),
            args: vec![
                format!(
                    "{}/llamafile",
                    std::env::current_dir().unwrap().to_string_lossy()
                )
                .as_str(),
                "--nobrowser",
            ]
            .into_iter()
            .map(|str| str.to_string())
            .chain(Into::<Vec<String>>::into(runner_args))
            .collect(),
        }
    }

    pub fn new_llama_server_bin(runner_args: RunnerArgs) -> Self {
        Self {
            pwd:  None,
            cmd:  "llama-server".to_string(),
            args: runner_args.into(),
        }
    }

    #[tracing::instrument(level = "debug")]
    pub async fn run(self) -> anyhow::Result<()> {
        let test = OsStr::new(self.cmd.as_str());

        let pwd = self
            .pwd
            .unwrap_or(std::env::current_dir()?.to_string_lossy().to_string());
        tracing::debug!(cmd=?test, ?pwd);

        let mut handle = Command::new(test)
            .env("HSA_OVERRIDE_GFX_VERSION", "11.0.0")
            .args(self.args.clone())
            .current_dir(pwd)
            .kill_on_drop(true)
            .spawn()?;

        handle.wait().await?.exit_ok()?;

        Ok(())
    }
}
