use axum::{
    body::Body,
    extract::State,
    http::{header, Request, Response, StatusCode, Uri},
    response::{IntoResponse, Response as AxumResponse},
};
use leptos::{provide_context, view};
use rust_embed::RustEmbed;

use crate::{app::App, server::AppState};

#[derive(RustEmbed)]
#[folder = "target/site"]
struct SiteAsset;

pub struct StaticFile<T>(pub T);

impl<T> IntoResponse for StaticFile<T>
where
    T: Into<String>,
{
    fn into_response(self) -> Response<Body> {
        let path = self.0.into();

        match SiteAsset::get(path.as_str()) {
            Some(content) => {
                let mime = mime_guess::from_path(path).first_or_octet_stream();
                ([(header::CONTENT_TYPE, mime.as_ref())], content.data).into_response()
            }
            None => (StatusCode::NOT_FOUND, "404 Not Found").into_response(),
        }
    }
}

pub async fn file_and_error_handler(
    uri: Uri,
    State(app_state): State<AppState>,
    req: Request<Body>,
) -> AxumResponse {
    let path = uri.path().trim_start_matches('/').to_string();
    let embed_res = StaticFile(path.clone()).into_response();
    if embed_res.status().is_success() {
        tracing::debug!(msg = "Found asset", ?path);
        embed_res
    } else {
        tracing::debug!(msg = "No Asset Found, Forwarding to leptos", ?path);
        let handler = leptos_axum::render_app_to_stream_with_context(
            app_state.leptos_options.to_owned(),
            move || {
                provide_context(app_state.pool.clone());
            },
            move || view! { <App/> },
        );
        handler(req).await.into_response()
    }
}
