mod logging;

use anyhow::anyhow;
use axum::{
    body::Body,
    http::{self, Request, Response},
    response::IntoResponse,
    routing::any,
    Router,
};
use futures;
use itertools::Itertools;
use reqwest::Client;
use serde::Deserialize;
use std::{collections::HashMap, net::SocketAddr, process::Stdio, sync::Arc};
use tokio::{
    net::TcpStream,
    process::{Child, Command},
    sync::Mutex,
    time::{sleep, Duration},
};
use tower_http::trace::{
    DefaultMakeSpan, DefaultOnEos, DefaultOnFailure, DefaultOnRequest, DefaultOnResponse,
    TraceLayer,
};
use tracing::Level;

use std::sync::Once;

pub static INIT: Once = Once::new();

pub fn initialize_logger() {
    INIT.call_once(|| {
        let env_filter = tracing_subscriber::EnvFilter::builder()
            .with_default_directive(tracing::level_filters::LevelFilter::INFO.into())
            .from_env_lossy();

        tracing_subscriber::fmt()
            .compact()
            .with_env_filter(env_filter)
            .init();
    });
}

#[derive(Clone, Debug, Deserialize)]
struct Config {
    hardware: Hardware,
    models: Vec<ModelConfig>,
}

impl Config {
    // TODO split up into raw deser config and "parsed"/"processed" config which always has a port
    fn pick_open_ports(self) -> Self {
        let mut config = self.clone();
        for model in &mut config.models {
            if model.internal_port.is_none() {
                model.internal_port = Some(
                    openport::pick_random_unused_port()
                        .expect(format!("No open port found for {:?}", model).as_str()),
                );
            }
        }
        config
    }
}

#[derive(Clone, Debug, Deserialize)]
struct Hardware {
    ram: String,
    vram: String,
}

#[derive(Clone, Debug, Deserialize)]
struct ModelConfig {
    #[allow(dead_code)]
    name: String,
    port: u16,
    internal_port: Option<u16>,
    env: HashMap<String, String>,
    args: HashMap<String, String>,
    vram_usage: String,
    ram_usage: String,
}

#[derive(Clone, Debug)]
struct LlamaInstance {
    config: ModelConfig,
    process: Arc<Mutex<Child>>,
    // busy: bool,
}

#[derive(Clone, Debug)]
struct SharedState {
    total_ram: u64,
    total_vram: u64,
    used_ram: u64,
    used_vram: u64,
    instances: HashMap<u16, LlamaInstance>,
}

#[tokio::main]
async fn main() {
    // TODO add autostart of models based on config
    // abstract starting logic out of handler for this to allow seperate calls to start
    // maybe add to SharedState & LLamaInstance ?

    initialize_logger();
    // Read and parse the YAML configuration
    let config_str = std::fs::read_to_string("config.yaml").expect("Failed to read config.yaml");
    let config: Config = serde_yaml::from_str::<Config>(&config_str).expect("Failed to parse config.yaml").pick_open_ports();

    // Parse hardware resources
    let total_ram = parse_size(&config.hardware.ram).expect("Invalid RAM size in config");
    let total_vram = parse_size(&config.hardware.vram).expect("Invalid VRAM size in config");

    // Initialize shared state
    let shared_state = Arc::new(Mutex::new(SharedState {
        total_ram,
        total_vram,
        used_ram: 0,
        used_vram: 0,
        instances: HashMap::new(),
    }));

    // For each model, set up an axum server listening on the specified port
    let mut handles = Vec::new();

    for model_config in config.models {
        let state = shared_state.clone();
        let model_config = model_config.clone();

        let handle = tokio::spawn(async move {
            let model_config = model_config.clone();

            let app = Router::new()
                .route(
                    "/",
                    any({
                        let state = state.clone();
                        let model_config = model_config.clone();
                        move |req| handle_request(req, model_config, state)
                    }),
                )
                .route(
                    "/*path",
                    any({
                        let state = state.clone();
                        let model_config = model_config.clone();
                        move |req| handle_request(req, model_config, state)
                    }),
                )
                .layer(
                    TraceLayer::new_for_http()
                        .make_span_with(DefaultMakeSpan::new().include_headers(true))
                        .on_request(DefaultOnRequest::new().level(Level::DEBUG))
                        .on_response(DefaultOnResponse::new().level(Level::TRACE))
                        .on_eos(DefaultOnEos::new().level(Level::DEBUG))
                        .on_failure(DefaultOnFailure::new().level(Level::ERROR)),
                );

            let addr = SocketAddr::from(([0, 0, 0, 0], model_config.port));

            tracing::info!(msg = "Listening", ?model_config);
            let listener = tokio::net::TcpListener::bind(&addr).await.unwrap();

            axum::serve(listener, app.into_make_service())
                .await
                .unwrap();
        });

        handles.push(handle);
    }

    futures::future::join_all(handles).await;
}

use thiserror::Error;

#[derive(Error, Debug)]
pub enum AppError {
    #[error("Axum Error")]
    AxumError(#[from] axum::Error),
    #[error("Axum Http Error")]
    AxumHttpError(#[from] axum::http::Error),
    #[error("Reqwest Error")]
    ReqwestError(#[from] reqwest::Error),
    #[error("Reqwest Middleware Error")]
    ReqwestMiddlewareError(#[from] reqwest_middleware::Error),
    #[error("Client Error")]
    ClientError(#[from] hyper::Error),
    #[error("Unknown error")]
    Unknown(#[from] anyhow::Error),
}

// Tell axum how to convert `AppError` into a response.
impl IntoResponse for AppError {
    fn into_response(self) -> axum::response::Response {
        tracing::error!("::AppError::into_response:: {:?}", self);
        (
            http::StatusCode::INTERNAL_SERVER_ERROR,
            format!("Something went wrong: {:?}", self),
        )
            .into_response()
    }
}

async fn handle_request(
    req: Request<Body>,
    model_config: ModelConfig,
    state: Arc<Mutex<SharedState>>,
    // ) -> Result<Response<Body>, anyhow::Error> {
) -> impl IntoResponse {
    let model_ram_usage = parse_size(&model_config.ram_usage).expect("Invalid ram_usage");
    let model_vram_usage = parse_size(&model_config.vram_usage).expect("Invalid vram_usage");

    let instance = {
        let mut state = state.lock().await;

        // Check if instance is running
        if let Some(instance) = state.instances.get_mut(&model_config.port) {
            // instance.busy = true;
            instance.to_owned()
        } else {
            // Check resources
            tracing::info!(msg = "Current state", ?state);
            if ((state.used_ram + model_ram_usage) > state.total_ram)
                || ((state.used_vram + model_vram_usage) > state.total_vram)
            {
                // Stop other instances
                let mut to_remove = Vec::new();
                // TODO Actual smart stopping logic
                // - search for smallest single model to stop to get enough room
                // - if not possible search for smallest number of models to stop with lowest
                //   amount of "overshot
                let instances_by_size =
                    state
                        .instances
                        .clone()
                        .into_iter()
                        .sorted_by(|(_, el_a), (_, el_b)| {
                            Ord::cmp(
                                &parse_size(el_b.config.vram_usage.as_str()),
                                &parse_size(el_a.config.vram_usage.as_str()),
                            )
                        });
                for (port, instance) in instances_by_size {
                    // if !instance.busy {
                    tracing::info!("Stopping instance on port {}", port);
                    let mut process = instance.process.lock().await;
                    process.kill().await.ok();
                    to_remove.push(port);
                    state.used_ram -= parse_size(&instance.config.ram_usage).unwrap_or(0);
                    state.used_vram -= parse_size(&instance.config.vram_usage).unwrap_or(0);
                    if state.used_ram + model_ram_usage <= state.total_ram
                        && state.used_vram + model_vram_usage <= state.total_vram
                    {
                        tracing::info!("Should have enough ram now");
                        break;
                    }
                    // }
                }
                for port in to_remove {
                    tracing::info!("Removing instance on port {}", port);
                    state.instances.remove(&port);
                }
            } else {
                tracing::info!("Already enough res free");
            }

            // Start new instance
            let args = model_config
                .args
                .iter()
                .flat_map(|(k, v)| {
                    if v == "true" {
                        vec![format!("--{}", k)]
                    } else {
                        vec![format!("--{}", k), v.clone()]
                    }
                })
                .collect::<Vec<_>>();

            let mut cmd = Command::new("llama-server");
            cmd.kill_on_drop(true);
            cmd.envs(model_config.env.clone());
            cmd.args(&args);
            // TODO use openport crate via pick_random_unused_port for determining these
            cmd.arg("--port");
            cmd.arg(format!("{}", model_config.internal_port.expect("Unexpected empty port, should've been picked")));
            cmd.stdout(Stdio::null()).stderr(Stdio::null()); // TODO save output and allow retrieval via api

            tracing::info!("Starting llama-server with {:?}", cmd);
            let process = Arc::new(Mutex::new(
                cmd.spawn().expect("Failed to start llama-server"),
            ));

            state.used_ram += model_ram_usage;
            state.used_vram += model_vram_usage;

            let instance = LlamaInstance {
                config: model_config.clone(),
                process,
                // busy: true,
            };
            sleep(Duration::from_millis(500)).await;
            state.instances.insert(model_config.port, instance.clone());

            instance
        }
    };

    // Wait for the instance to be ready
    is_llama_instance_running(&instance).await?;
    wait_for_port(model_config.internal_port.expect("Unexpected empty port, should've been picked")).await?;

    // Proxy the request
    let retry_policy = reqwest_retry::policies::ExponentialBackoff::builder()
        .retry_bounds(Duration::from_secs(1), Duration::from_secs(8))
        .jitter(reqwest_retry::Jitter::None)
        .base(2)
        .build_with_max_retries(8);

    let client = reqwest_middleware::ClientBuilder::new(Client::new())
        .with(reqwest_retry::RetryTransientMiddleware::new_with_policy(
            retry_policy,
        ))
        .build();

    let uri = format!(
        "http://127.0.0.1:{}{}",
        model_config.internal_port.expect("Unexpected empty port, should've been picked"),
        req.uri().path_and_query().map(|x| x.as_str()).unwrap_or("")
    );

    let mut request_builder = client.request(req.method().clone(), &uri);
    // let mut request_builder = reqwest::RequestBuilder::from_parts(client, req.method().clone())

    for (name, value) in req.headers() {
        request_builder = request_builder.header(name, value);
    }

    let body_bytes = axum::body::to_bytes(
        req.into_body(),
        parse_size("1G").unwrap().try_into().unwrap(),
    )
    .await?;
    let request = request_builder.body(body_bytes).build()?;
    // tracing::info!("Proxying request to {}", uri);
    let response = client.execute(request).await?;

    // tracing::info!("Received response from {}", uri);
    let mut builder = Response::builder().status(response.status());
    for (name, value) in response.headers() {
        builder = builder.header(name, value);
    }

    // let bytes = response.bytes().await?;

    let byte_stream = response.bytes_stream();

    let body = axum::body::Body::from_stream(byte_stream);

    tracing::debug!("streaming response on port: {}", model_config.port);
    let response = builder.body(body)?;
    Ok::<axum::http::Response<Body>, AppError>(response)
}

async fn wait_for_port(port: u16) -> Result<(), anyhow::Error> {
    for _ in 0..10 {
        if TcpStream::connect(("127.0.0.1", port)).await.is_ok() {
            return Ok(());
        }
        sleep(Duration::from_millis(500)).await;
    }
    Err(anyhow!("Timeout waiting for port"))
}

// reimplement wait_for_port using a LlamaInstance, also check if the process is still running
async fn is_llama_instance_running(instance: &LlamaInstance) -> Result<(), anyhow::Error> {
    match instance.process.to_owned().lock_owned().await.try_wait()? {
        Some(exit_status) => {
            let msg = "Llama instance exited";
            tracing::error!(msg, ?exit_status);
            Err(anyhow!(msg))
        }

        None => Ok(()),
    }
}

fn parse_size(size_str: &str) -> Option<u64> {
    let mut num = String::new();
    let mut unit = String::new();

    for c in size_str.chars() {
        if c.is_digit(10) || c == '.' {
            num.push(c);
        } else {
            unit.push(c);
        }
    }

    let num: f64 = num.parse().ok()?;

    let multiplier = match unit.to_lowercase().as_str() {
        "g" | "gb" => 1024 * 1024 * 1024,
        "m" | "mb" => 1024 * 1024,
        "k" | "kb" => 1024,
        _ => panic!("Invalid Size"),
    };

    let res = (num * multiplier as f64) as u64;
    Some(res)
}
